package edu.washington.cs.radishsim;

import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.NavigableSet;
import java.util.TreeSet;

class AsyncCheck implements Comparable<AsyncCheck> {

   final long             start;
   final CpuId            owner;
   final List<DataAccess> accesses;

   AsyncCheck(CpuId c, long start, List<DataAccess> acc) {
      this.start = start;
      this.owner = c;
      this.accesses = acc;
   }

   /** Order Checks by their start time. */
   @Override
   public int compareTo(AsyncCheck c) {
      if ( !(c instanceof AsyncCheck) ) throw new ClassCastException();
      if ( this.start < c.start ) return -1;
      else if ( this.start == c.start ) return 0;
      else return 1;
   }

   @Override
   public String toString() {
      return "[" + this.start + "]@" + this.owner.toString();
   }
}

public class CheckerProcessor<Line extends RadishLine> extends Processor<Line> {

   /** Stores the list of pending checks, ordered by start time. */
   private final NavigableSet<AsyncCheck> pendingChecks = new TreeSet<AsyncCheck>();

   /**
    * The most recent time at which some check from each cpuid completed. If there are no pending
    * checks for cpuid C, then checkCompletionTime[C] is the time at which C's most-recent check
    * completed.
    */
   private final long[]                   checkCompletionTime;

   public CheckerProcessor(Machine.MachineParams<Line> args, Machine<Line> machine, CpuId cpuid,
                           SoftwareMetadata globalTab, Map<Long, SyncObject> syncObjectMap,
                           Processor<Line>[] processors,
                           Map<UnownedMetadataByteAddress, InHardwareStatus> inHwStatus,
                           Map<LineAddress, Integer> varmap, CheckerProcessor<Line>[] checkers) {
      super(args, machine, cpuid, globalTab, syncObjectMap, processors, inHwStatus, varmap,
            checkers);
      checkCompletionTime = new long[processors.length];
   }

   /** Return the current execution time at this checker proc. */
   public long currentTime() {
      return stats.pc_ExecDrivenCycleCount.get();
   }

   /** Schedule a check at this checker proc. */
   public void scheduleCheck(CpuId source, long desiredStartTime, List<DataAccess> checkAccesses) {
      // check gets next available slot
      long start = Math.max(desiredStartTime, currentTime());
      pendingChecks.add(new AsyncCheck(source, start, checkAccesses));
   }

   /**
    * Executes all the pending checks from source, and any checks from other processors that were
    * enqueued before the most-recent check from source.
    * 
    * @return the time when the last check from source completes
    */
   public long flushChecks(CpuId source) {

      // count # checks from source
      int sourceChecks = 0;
      for ( AsyncCheck ac : pendingChecks ) {
         if ( ac.owner.equals(source) ) sourceChecks++;
      }

      if ( sourceChecks == 0 ) { return checkCompletionTime[source.get()]; }

      for ( Iterator<AsyncCheck> i = pendingChecks.iterator(); i.hasNext(); ) {
         AsyncCheck ac = i.next();
         i.remove();
         if ( ac.owner.equals(source) ) sourceChecks--;

         // start next check as soon as possible
         long checkStart = Math.max(ac.start, currentTime());
         stats.pc_ExecDrivenCycleCount.set(checkStart);

         // execute next check's accesses, which advances our current time
         for ( DataAccess da : ac.accesses ) {
            if ( da.read() ) this.read(da);
            else this.write(da);
         }
         
         stats.pc_ExecDrivenCycleCount.incr(MemorySystemConstants.FASTTRACK_CHECK_INSN_COST);

         checkCompletionTime[ac.owner.get()] = currentTime();

         if ( sourceChecks == 0 ) break;
      }

      return currentTime();
   }

}
